function [x,u,J,converg,iterat] = NLloop_Pareto(fname,NP,NU,x,xSm1,u,z,theta,x_target,u_target,Wxx,Wtxx,Wuu,Wtuu,alpha,muPareto,eps,stochpar,T,S,S_subgame,TE,NLmaxit,NLtoler,Ssolver,GSmaxit,Nmaxit,SStoler,disp_err,MC)
%running the nonlinearity loop for cooperative pareto solution until the algorithm converges, or the NLmaxit
%is achieved, or the results worsen in 3 runs in a row.
%in each iteration run backward according to equations (89) and (98) and
%then forward.
%After that the convergence of the algortihm is checked and the
%same procedure restarted again (if there is no convergence)

n=size(x,2);
m=size(u,2);

iterat = 0;
converg = 0;
diverg = 0;
divergence_info_on = 1; %if = 1 report (temporal) divergence

while ((converg < (n + m) * (T - S_subgame + 1)) && (iterat < NLmaxit) && diverg < 3)
    iterat = iterat + 1;
    if iterat==1
        Jiter_start = calculateJ(NP,x,u,x_target,u_target,Wxx,Wuu,Wtxx,Wtuu,alpha,T,S,TE,0,0,0,0,0);
    else
        Jiter_start = J;
    end
    J = 0;
    xold = x;
    uold = u;
    
    H=zeros(n,n);
    h=zeros(n,1);
    A=zeros(n,n);
    B=zeros(n,m);
    G=zeros(m,n,T - S_subgame + 1);
    g=zeros(m,T - S_subgame + 1);
    K=zeros(n,n,T - S_subgame + 1);
    k=zeros(n,T - S_subgame + 1);
    
    for t=T:-1:S_subgame
        alpha_t=1/(alpha^(t-1)); %discount factor in t
        %alpha_tm1=1/(alpha^(t-2)); %discount factor in t-1
        if t==TE %check for special time periods with different weights Wtxx, Wtuu
            Q=Wtxx;
            R=Wtuu;
        else
            Q=Wxx;
            R=Wuu;
        end
        x0t = x(t,:);
        if t==1
            x0tm1 = xSm1;
        else
            x0tm1 = x(t-1,:);
        end
        u0t = u(t,:);
        z0t = z(t,:);
        eps = zeros(n,1);
        %linearize system
        [A(:,:),B(:,:),c] = linearization(fname,x0t,x0tm1,u0t,z0t,theta,eps);
        
        Qt=zeros(n,n);
        qt=zeros(n,1);
        Rj=zeros(m,m);
        rj=zeros(m,1);
%         Rjp1=zeros(m,m);
%         rjp1=zeros(m,1);
        
        for player=1:NP %equations 89-92
            Qt=Qt+muPareto(player)*Q(:,:,player)*alpha_t;
            qt=qt+muPareto(player)*Q(:,:,player)*alpha_t*x_target(t,:,player)';
            Rj(:,:)=Rj(:,:)+muPareto(player)*R(:,:,player)*alpha_t;
            rj(:,:)=rj(:,1)+muPareto(player)*R(:,:,player)*alpha_t*u_target(t,:,player)';
        end
        %save Rj and rj from t+1
        Rjp1=Rj;
        rjp1=rj;
        
        if t==T
            H(:,:)=Qt;
            h(:,1)=qt;
        else
            Htm1=zeros(n,n); %calculate Htm1 and htm1 (equations 93, 94)
            htm1=zeros(n,1);
            GRG = G(:,:,t+1)'*Rjp1(:,:)*G(:,:,t+1);
            GrmRg = G(:,:,t+1)'*(rjp1(:,1)-Rjp1(:,:)*g(:,t+1));
            Htm1(:,:)=Qt(:,:) + (K(:,:,t+1)')*H(:,:)*K(:,:,t+1) + GRG;
            htm1(:,1)=qt(:,:) - (K(:,:,t+1)')*(H(:,:)*k(:,t+1)-h(:,1)) + GrmRg;
            H=Htm1;
            h=htm1;
        end
    
        %Gauss-Seidel to solve equation 97
        iter = 0; Gt=zeros(m,n); Gt_old = Gt+1;
        while (norm(Gt-Gt_old) > 0.00001 &&  iter < GSmaxit) %
            Gt_old = Gt;
            nu=1;
            for player=1:NP
                nupi=nu+NU(player)-1; %nu:nupi control variables of player i
                factor_sumj=zeros(n,n);
                nuj=1;
                for player_j=1:NP
                    nujpj=nuj+NU(player_j)-1;
                    if player_j ~= player
                        factor_sumj=factor_sumj+B(:,nuj:nujpj)*Gt(nuj:nujpj,:);
                    end
                    nuj=nuj+NU(player_j);
                end
                factor_b = -B(:,nu:nupi)'*H(:,:)*A - B(:,nu:nupi)'*H(:,:)*factor_sumj;
                factor_A = Rj(nu:nupi,nu:nupi) + B(:,nu:nupi)'*H(:,:)*B(:,nu:nupi);
                Gt(nu:nupi,:)=factor_A\factor_b;
                nu=nu+NU(player);
            end
            iter= iter + 1;
        end % while
        if iter == GSmaxit
            disp(['exit for gauss-seidel because maximal number of runs (: ' num2str(GSmaxit) ') was reached in time period ' num2str(t)])
        end
        G(:,:,t)=Gt;
        
        %Gauss-Seidel to solve equation 98
        iter = 0; gt=zeros(m,1); gt_old = gt+1;
        while (norm(gt-gt_old) > 0.00001 &&  iter < GSmaxit) %
            gt_old = gt;
            nu=1;
            for player=1:NP
                nupi=nu+NU(player)-1; %nu:nupi control variables of player i
                factor_sumj=zeros(n,1);
                nuj=1;
                for player_j=1:NP
                    nujpj=nuj+NU(player_j)-1;
                    if player_j ~= player
                        factor_sumj=factor_sumj+B(:,nuj:nujpj)*gt(nuj:nujpj,:);
                    end
                    nuj=nuj+NU(player_j);
                end
                factor_b = B(:,nu:nupi)'*h(:,1) + rj(nu:nupi,1) -B(:,nu:nupi)'*H(:,:)*c - B(:,nu:nupi)'*H(:,:)*factor_sumj;
                factor_A = Rj(nu:nupi,nu:nupi) + B(:,nu:nupi)'*H(:,:)*B(:,nu:nupi);
                gt(nu:nupi,:)=factor_A\factor_b;
                nu=nu+NU(player);
            end
            iter= iter + 1;
        end % while
        if iter == GSmaxit
            disp(['exit for gauss-seidel because maximal number of runs (: ' num2str(GSmaxit) ') was reached in time period ' num2str(t)])
        end
        g(:,t)=gt;
        
        
        %calculate K
        K(:,:,t)=A+B*G(:,:,t);
        k(:,t)=c+B*g(:,t);
        
    end
    
    
    %'Forward loop': Compute new values for x and u
    
% linear solution

%     while t <= T - S + 1
%         if t==1
%             x(t,:)=K(:,:,t)*xSm1' + k(:,t);
%             u(t,:)=G(:,:,t)*xSm1' + g(:,t);
%         else
%             x(t,:)=K(:,:,t)*x(t-1,:)' + k(:,t);
%             u(t,:)=G(:,:,t)*x(t-1,:)' + g(:,t);
%         end
%         
%         t = t + 1;
%     end
    

% nonlinear solution 
    while t <= T - S + 1
        if t==1
            x0_t = xSm1;                        % Starting guess
            xl_t = xSm1;
            u(t,:)=G(:,:,t)*xSm1' + g(:,t);
        else
            x0_t = x(t-1,:);                    % Starting guess
            xl_t = x(t-1,:);
            u(t,:)=G(:,:,t)*x(t-1,:)' + g(:,t);
        end
        u_t = u(t,:);
        z_t = z(t,:);
        [x(t,:),res,iter,fval]=fsolution(Ssolver,fname,x0_t,xl_t,u_t,z_t,theta,eps,Nmaxit,GSmaxit,SStoler,disp_err,t,0);
        
        t = t + 1;
    end
    
    %check the convergency
    converg = 0;
    xdiff = abs((x-xold)./xold);
    udiff = abs((u-uold)./uold);
    for i=1:(T-S_subgame+1)
        for j=1:n
            if xdiff(i,j) <= NLtoler
                converg = converg + 1;
            end
        end
        for jj=1:m
            if udiff(i,jj) <= NLtoler
                converg = converg + 1;
            end
        end
    end
    
    %J = calculateJ(NP,x,u,x_target,u_target,Q,R,Wtxx,Wtuu,alpha,T,S,TE,0,0,0,0,0);
    J = calculateJ(NP,x,u,x_target,u_target,Wxx,Wuu,Wtxx,Wtuu,alpha,T,S,TE,0,0,0,0,0);
    Jiter_diff=J-Jiter_start;
    if (Jiter_diff(NP+1))>0.001 && divergence_info_on == 1
        disp(['(temporal) divergence in Pareto strategy: iterat = ' num2str(iterat) '; Jiter_diff = ' num2str(Jiter_diff)])
    end
end