function [x,u,J,converg,iterat] = NLloop_StackFB(fname,NP,NU,x,xSm1,u,z,theta,x_target,u_target,Wxx,Wtxx,Wuu,Wtuu,alpha,eps,stochpar,T,S,TE,NLmaxit,NLtoler,Ssolver,GSmaxit,Nmaxit,SStoler,disp_err,MC)
%running the nonlinearity loop for Stackelberg feedback solution until the algorithm converges, or the NLmaxit
%is achieved, or the results worsen in 3 runs in a row.
%in each iteration run backward according to equations (116) and (131) and
%then forward.
%After that the convergence of the algortihm is checked and the
%same procedure restarted again (if there is no convergence)

n=size(x,2);
m=size(u,2);

iterat = 0;
converg = 0;
diverg = 0;
divergence_info_on = 1; %if = 1 report (temporal) divergence

while ((converg < (n + m) * (T - S + 1)) && (iterat < NLmaxit) && diverg < 3)
    iterat = iterat + 1;
    if iterat==1
        Jiter_start = calculateJ(NP,x,u,x_target,u_target,Wxx,Wuu,Wtxx,Wtuu,alpha,T,S,TE,0,0,0,0,0);
    else
        Jiter_start = J;
    end
    J = 0;
    xold = x;
    uold = u;
    
    H=zeros(n,n,NP);
    h=zeros(n,NP);
    A=zeros(n,n);
    B=zeros(n,m);
    G=zeros(m,n,T - S + 1);
    g=zeros(m,T - S + 1);
    K=zeros(n,n,T - S + 1);
    k=zeros(n,T - S + 1);
    
    for t=T:-1:S
        alpha_t=1/(alpha^(t-1)); %discount factor in t
        alpha_tm1=1/(alpha^(t-2)); %discount factor in t-1
        if t==TE %check for special time periods with different weights Wtxx, Wtuu
            Q=Wtxx;
            R=Wtuu;
        else
            Q=Wxx;
            R=Wuu;
        end
        x0t = x(t,:);
        if t==1
            x0tm1 = xSm1;
        else
            x0tm1 = x(t-1,:);
        end
        u0t = u(t,:);
        z0t = z(t,:);
        eps = zeros(n,1);
        %linearize system
        [A(:,:),B(:,:),c] = linearization(fname,x0t,x0tm1,u0t,z0t,theta,eps);
        
        if t==T
            %nu=1;
            for player=1:NP
                %nupi=nu+NU(player)-1; %nu:nupi control variables of player i
                H(:,:,player)=Q(:,:,player)*alpha_t;
                h(:,player)=Q(:,:,player)*alpha_t*x_target(t,:,player)';
                %nu=nu+NU(player);
            end
        end
        
        D=zeros(1,1);
        nu=NU(1)+1;
        for player=2:NP %equation 118
            nupi=nu+NU(player)-1; %nu:nupi control variables of player i
            D(nu:nupi,nu:nupi)=R(nu:nupi,nu:nupi,player)*alpha_t+B(:,nu:nupi)'*H(:,:,player)*B(:,nu:nupi);
            nu=nu+NU(player);
        end
        
        
        %Gauss-Seidel to solve equation 116
        iter = 0; Psi=zeros(m,1); Psi_old = Psi+1;
        while (norm(Psi-Psi_old) > 0.00001 &&  iter < GSmaxit) %
            Psi_old = Psi;
            nu=NU(1)+1;
            for player=2:NP
                nupi=nu+NU(player)-1; %nu:nupi control variables of player i
                factor_sumj=0;
                nuj=NU(1)+1;
                for player_j=2:NP
                    nujpj=nuj+NU(player_j)-1;
                    if player_j ~= player
                        factor_sumj=factor_sumj+B(:,nu:nupi)'*H(:,:,player)*B(:,nuj:nujpj)*Psi(nuj:nujpj,:);
                    end
                    nuj=nuj+NU(player_j);
                end
                factor_b=-B(:,nu:nupi)'*H(:,:,player)*B(:,1)-factor_sumj;
                Psi(nu:nupi,:)=D(nu:nupi,nu:nupi)\factor_b;
                nu=nu+NU(player);
            end
            iter= iter + 1;
        end % while
        if iter == GSmaxit
            disp(['exit for gauss-seidel because maximal number of runs (: ' num2str(GSmaxit) ') was reached in time period ' num2str(t)])
        end
        
        Bquer = B(:,1:NU(1));
        nu=NU(1)+1;
        for player=2:NP %equation 123
            nupi=nu+NU(player)-1; %nu:nupi control variables of player i
            Bquer = Bquer + B(:,nu:nupi)*Psi(nu:nupi);
            nu=nu+NU(player);
        end
        Lambda=R(1:NU(1),1:NU(1),1)*alpha_t + (Bquer')*H(:,:,1)*Bquer;
        Dquer=zeros(1,m);
        vquer = zeros(1,NU(1));
        nu=NU(1)+1;
        for player=2:NP %equations 122, 128, 129
            nupi=nu+NU(player)-1; %nu:nupi control variables of player i
            Lambda = Lambda + (Psi(nu:nupi)')*R(nu:nupi,nu:nupi,1)*alpha_t*Psi(nu:nupi);
            Dquer(1,nu:nupi) = (Psi(nu:nupi)')*R(nu:nupi,nu:nupi,1)*alpha_t + (Bquer')*H(:,:,1)*B(:,nu:nupi);
            vquer = vquer + (Psi(nu:nupi)')*(B(:,nu:nupi)'*H(:,:,1)*c-B(:,nu:nupi)'*h(:,1) - R(nu:nupi,nu:nupi,1)*alpha_t*u_target(t,nu:nupi,1)');
            nu=nu+NU(player);
        end
        
        
        %Gauss-Seidel to solve equation 119
        iter = 0; W=zeros(m,n); W_old = W+1;
        while (norm(W-W_old) > 0.00001 &&  iter < GSmaxit) %
            W_old = W;
            nu=NU(1)+1;
            for player=2:NP
                nupi=nu+NU(player)-1; %nu:nupi control variables of player i
                factor_sumj=A;
                nuj=NU(1)+1;
                for player_j=2:NP
                    nujpj=nuj+NU(player_j)-1;
                    if player_j ~= player
                        factor_sumj=factor_sumj+B(:,nuj:nujpj)*W(nuj:nujpj,:);
                    end
                    nuj=nuj+NU(player_j);
                end
                factor_b=-B(:,nu:nupi)'*H(:,:,player)*factor_sumj;
                W(nu:nupi,:)=D(nu:nupi,nu:nupi)\factor_b;
                nu=nu+NU(player);
            end
            iter= iter + 1;
        end % while
        if iter == GSmaxit
            disp(['exit for gauss-seidel because maximal number of runs (: ' num2str(GSmaxit) ') was reached in time period ' num2str(t)])
        end
        
        
        %Gauss-Seidel to solve equation 120
        iter = 0; w=zeros(m,1); w_old = w+1;
        while (norm(w-w_old) > 0.00001 &&  iter < GSmaxit) %
            w_old = w;
            nu=NU(1)+1;
            for player=2:NP
                nupi=nu+NU(player)-1; %nu:nupi control variables of player i
                factor_sumj=zeros(n,1);
                nuj=NU(1)+1;
                for player_j=2:NP
                    nujpj=nuj+NU(player_j)-1;
                    if player_j ~= player
                        factor_sumj=factor_sumj+B(:,nuj:nujpj)*w(nuj:nujpj,:);
                    end
                    nuj=nuj+NU(player_j);
                end
                factor_b=-B(:,nu:nupi)'*(H(:,:,player)*c-h(:,player)) + R(nu:nupi,nu:nupi,player)*alpha_t*u_target(t,nu:nupi,player)'-B(:,nu:nupi)'*H(:,:,player)*factor_sumj;
                w(nu:nupi,:)=D(nu:nupi,nu:nupi)\factor_b;
                nu=nu+NU(player);
            end
            iter= iter + 1;
        end % while
        if iter == GSmaxit
            disp(['exit for gauss-seidel because maximal number of runs (: ' num2str(GSmaxit) ') was reached in time period ' num2str(t)])
        end
        
        
        G(1:NU(1),:,t) = - inv(Lambda)*(Bquer'*H(:,:,1)*A + Dquer(1,NU(1)+1:m)*W(NU(1)+1:m,:));
        g(1:NU(1),t) = - inv(Lambda)*(B(:,1:NU(1))'*(H(:,:,1)*c-h(:,1))-R(1,1,1)*alpha_t*u_target(t,1:NU(1),1) + vquer + Dquer(1,NU(1)+1:m)*w(NU(1)+1:m,:));
        nu=NU(1)+1;
        for player=2:NP
            nupi=nu+NU(player)-1; %nu:nupi control variables of player i
            G(nu:nupi,:,t) = W(nu:nupi,:) + Psi(nu:nupi,:)*G(1:NU(1),:,t);
            g(nu:nupi,t) = w(nu:nupi,:) + Psi(nu:nupi,:)*g(1:NU(1),t);
            nu=nu+NU(player);
        end
        
        %calculate K
        K(:,:,t)=A+B*G(:,:,t);
        k(:,t)=c+B*g(:,t);
        
        Htm1=zeros(n,n,NP); %calculate Ht and ht for time period t-1
        htm1=zeros(n,NP);
        if t>S
            %nu=1;
            for player=1:NP
                %nupi=nu+NU(player)-1; %nu:nupi control variables of player i
                Htm1(:,:,player)=Q(:,:,player)*alpha_tm1+(K(:,:,t)')*H(:,:,player)*K(:,:,t) + (G(:,:,t)')*R(:,:,player)*alpha_t*G(:,:,t);
                htm1(:,player)=Q(:,:,player)*alpha_tm1*x_target(t-1,:,player)' - (K(:,:,t)')*(H(:,:,player)*k(:,t)-h(:,player)) + (G(:,:,t)')*R(:,:,player)*alpha_t*(u_target(t,:,player)'-g(:,t));
                %nu=nu+NU(player);
            end
            H=Htm1;
            h=htm1;
        end
        
    end
    
    
    %'Forward loop': Compute new values for x and u
    
% linear solution

%     while t <= T - S + 1
%         if t==1
%             x(t,:)=K(:,:,t)*xSm1' + k(:,t);
%             u(t,:)=G(:,:,t)*xSm1' + g(:,t);
%         else
%             x(t,:)=K(:,:,t)*x(t-1,:)' + k(:,t);
%             u(t,:)=G(:,:,t)*x(t-1,:)' + g(:,t);
%         end
%         
%         t = t + 1;
%     end
    
% nonlinear solution

    while t <= T - S + 1
        if t==1
            x0_t = xSm1;                        % Starting guess
            xl_t = xSm1;
            u(t,:)=G(:,:,t)*xSm1' + g(:,t);
        else
            x0_t = x(t-1,:);                    % Starting guess
            xl_t = x(t-1,:);
            u(t,:)=G(:,:,t)*x(t-1,:)' + g(:,t);
        end
        u_t = u(t,:);
        z_t = z(t,:);
        [x(t,:),res,iter,fval]=fsolution(Ssolver,fname,x0_t,xl_t,u_t,z_t,theta,eps,Nmaxit,GSmaxit,SStoler,disp_err,t,0);
        
        t = t + 1;
    end
    
    %check the convergency
    converg = 0;
    xdiff = abs((x-xold)./xold);
    udiff = abs((u-uold)./uold);
    for i=1:(T-S+1)
        for j=1:n
            if xdiff(i,j) <= NLtoler
                converg = converg + 1;
            end
        end
        for jj=1:m
            if udiff(i,jj) <= NLtoler
                converg = converg + 1;
            end
        end
    end
    
    %J = calculateJ(NP,x,u,x_target,u_target,Q,R,Wtxx,Wtuu,alpha,T,S,TE,0,0,0,0,0);
    J = calculateJ(NP,x,u,x_target,u_target,Wxx,Wuu,Wtxx,Wtuu,alpha,T,S,TE,0,0,0,0,0);
    Jiter_diff=J-Jiter_start;
    if (Jiter_diff(NP+1))>0 && divergence_info_on == 1
        disp(['(temporal) divergence in StackFB strategy: iterat = ' num2str(iterat) '; Jiter_diff = ' num2str(Jiter_diff)])
    end
end