function [x,res,iter,fval] = fsolution(Ssolver,fname,x0_t,xl_t,u_t,z_t,Etheta,eps,Nmaxit,GSmaxit,SStoler,disp_err,t,MC)
%calculate the nonlinear function f at given state
res=0;
iter=0;
fval=0;
if strcmpi(Ssolver, 'newton') == 1 %newton line search
    [x,res,iter] = newton(fname,x0_t,xl_t,u_t,z_t,Etheta,eps,Nmaxit,SStoler);
elseif strcmpi(Ssolver, 'gauss-seidel') == 1  %gauss-seidel
    [x,res,iter] = gs(fname,x0_t,xl_t,u_t,z_t,Etheta,eps,GSmaxit,SStoler);
    if iter == GSmaxit
        disp(['exit for gauss-seidel caused by max. allowed number of iterations in (t=' num2str(t) ', MC=' num2str(MC) ')'])
    end
elseif strcmpi(Ssolver, 'fsolve') == 1 %matlab function fsolve
    switch fname
        case {'slovnl'}
            f = @(x0_t)slovnl(x0_t,xl_t,u_t,z_t,Etheta,eps,1); %anonymous function (several parameters)
        case {'slopol4'}
            f = @(x0_t)slopol4(x0_t,xl_t,u_t,z_t,Etheta,eps,1);
        case {'slopol8'}
            f = @(x0_t)slopol8(x0_t,xl_t,u_t,z_t,Etheta,eps,1);
        case {'slovl'}
            f = @(x0_t)slovl(x0_t,xl_t,u_t,z_t,Etheta,eps,1);
        case {'mumod1'}
            f = @(x0_t)mumod1(x0_t,xl_t,u_t,z_t,Etheta,eps,1);
        case {'mumod1_1'}
            f = @(x0_t)mumod1_1(x0_t,xl_t,u_t,z_t,Etheta,eps,1);
        case {'cons_game'}
            f = @(x0_t)cons_game(x0_t,xl_t,u_t,z_t,Etheta,eps,1);
    end
    options = optimset('Display','off');
    [x,fval,exitflag] = fsolve(f,x0_t,options);    % Invoke optimizer
    if strcmpi(disp_err, 'on')
        disp(['exit flag for fsolve: ' num2str(exitflag) ' (t=' num2str(t) ', MC=' num2str(MC) ')'])
    end
    if strcmpi(disp_err, 'error')
        if exitflag < 1
            disp(['exit flag for fsolve: ' num2str(exitflag) ' (t=' num2str(t) ', MC=' num2str(MC) ')'])
            res = 1;
        end
    end
end